"""Deals with version numbers through literate regular expressions."""

import codecs
import re
from bag.text import content_of
from grimace import RE  # https://github.com/benlast/grimace/wiki/Documentation


VERSION_NUMBER = str(
    RE().one_or_more.digits.dot.one_or_more.digits.then.any_number_of.any_of(
        "0123456789-.()abcdefghijklmnopqrstuvwxyz"
    )
)
VERSION_NUMBER_RE = re.compile(VERSION_NUMBER)
VERSION_VALIDATOR = "^" + VERSION_NUMBER + "$"
VERSION_VALIDATOR_RE = re.compile(VERSION_VALIDATOR)


def error_in_version(val, allow_dev=True):
    """Validate a version number. Returns None if valid."""
    if not VERSION_VALIDATOR_RE.match(val):
        return '"{0}" is not a valid version number.'.format(val)
    if not allow_dev:
        for x in ("svn", "dev", "("):
            if x in val:
                return '"{0}" is a development version number.'.format(val)


SOME_QUOTE = r'["\']'  # single or double quote
QUOTED_VERSION = SOME_QUOTE + "(" + VERSION_NUMBER + ")" + SOME_QUOTE


def version_in_python_source(text, replace=None, keyword="version"):
    """Given a string, finds or replaces the version number in it.

    If ``replace`` is None, returns the version number found in ``text``.
    Else, returns an updated ``text`` (with the version number specified in
    the ``replace`` argument).
    """
    PYTHON_VERSION_LINE = str(
        RE()
        .zero_or_more.whitespace.between(0, 2)
        .underscore.then(keyword)
        .then.between(0, 2)
        .underscore.then.zero_or_more.whitespace.then("=")
        .then.zero_or_more.whitespace.then.regex(QUOTED_VERSION)
        .then.zero_or_more.whitespace
    )
    PYTHON_VERSION_LINE_RE = re.compile(PYTHON_VERSION_LINE)
    match = PYTHON_VERSION_LINE_RE.search(text)
    assert match, "Could not find version number in Python source code."
    version = match.groups()[0]
    if replace:
        alist = []
        found = False
        for line in text.splitlines(True):  # True keeps line endings
            match = PYTHON_VERSION_LINE_RE.match(line)
            if not found and match:
                found = True
                line = line.replace(version, replace)
            alist.append(line)
        assert found, "Could not replace version number in Python source code."
        return "".join(alist)
    else:
        return version


def version_in_python_source_file(
    path, replace=None, keyword="version", encoding="utf-8"
):
    """Given a file path, returns or replaces the version number in it.

    If replace is None, returns the version number found in file ``path``.
    Else, replaces the version number with the one specified in ``replace``.
    """
    ret = version_in_python_source(
        content_of(path, encoding=encoding), replace=replace, keyword=keyword
    )
    if replace:
        with codecs.open(path, "w", encoding=encoding) as stream:
            stream.write(ret)
        # Test the change we just did by reading the file we just wrote
        assert replace == version_in_python_source_file(
            path, keyword=keyword, encoding=encoding
        )
    else:
        return ret



"""Release steps involving git."""

from . import ReleaseStep, CommandStep, StopRelease

__all__ = (
    "EnsureGitClean",
    "EnsureGitBranch",
    "GitCommitVersionNumber",
    "GitTag",
    "GitPush",
    "GitPushTags",
)


class EnsureGitClean(ReleaseStep):
    """Ensures git state is appropriate to start a release.

    Meaning we are NOT in a detached head and
    there are no uncommitted changes in tracked files.
    """

    ERROR_CODE = 52

    def __call__(self):
        head = self._execute_or_complain("git symbolic-ref --quiet HEAD")
        # This returns something like 'refs/heads/master' or nothing.
        # Nothing is bad, indicating a detached head: likely a tag checkout
        if not head:
            raise StopRelease("Wait, are you on a detached head?")
        changes = self._execute_or_complain(
            "git status --short --untracked-files=no"
        )  # sets self.success
        if changes:
            raise StopRelease("There are uncommitted changes in tracked files.")
        self._succeed()


class EnsureGitBranch(ReleaseStep):
    """I must be in the branch specified in config."""

    ERROR_CODE = 51

    def _get_current_branch(self):
        branch = self._execute_or_complain("git rev-parse --abbrev-ref HEAD")
        return None if branch == "HEAD" else branch

    def __call__(self):
        required = self.config.get("branch", "master")
        branch = self._get_current_branch()
        if branch != required:
            raise StopRelease(
                'You are in branch "{0}", but should be '
                'in branch "{1}" in order to make a release.'.format(branch, required)
            )
        self._succeed()


class GitCommitVersionNumber(ReleaseStep):
    """Creates a git commit with only one alteration: the new version number.

    Can rollback().
    """

    COMMAND = 'git commit -a -m "{0}"'
    ERROR_CODE = 53

    def __init__(self, which="the_version", msg="Version {0}", stop_on_failure=True):
        assert which in ("the_version", "future_version")
        self.which = which
        self.msg = msg.replace('"', '"')  # escape quotes for shell
        self.stop_on_failure = stop_on_failure

    def __call__(self):
        msg = self.msg.format(getattr(self.releaser, self.which))
        self._execute_or_complain(self.COMMAND.format(msg))  # sets success

    def rollback(self):
        self._execute_or_complain("git reset --hard HEAD^")


class GitTag(ReleaseStep):
    """Tags the current git commit with the new version number.

    Can rollback().
    """

    COMMAND = 'git tag -a v{0} -m "Version {0}"'
    ERROR_CODE = 54
    stop_on_failure = False

    def __call__(self):
        version = self.releaser.the_version
        if version is None:
            self.log.warning(
                "Skipping the GitTag step. It can only run AFTER "
                "some other step sets *the_version* on the releaser."
            )
            return
        self._execute_or_complain(self.COMMAND.format(version))  # sets success
        if self.success:
            self.releaser.created_tags.append(version)

    def rollback(self):
        self._execute_or_complain('git tag -d "v{0}"'.format(self.releaser.the_version))


class GitPush(ReleaseStep):
    """Does ``git push``. This critical step has no rollback."""

    COMMAND = "git push"
    ERROR_CODE = 55
    stop_on_failure = False  # It should be easy to 'git push' afterwards
    no_rollback = (
        "One should never try to undo a git push. Really.\n"
        "This release process went far -- the push succeeded -- but\n"
        "something else went wrong after the push. I figure it will\n"
        "be easier for you to finish the release process manually,\n"
        "so I will NOT roll back any of the steps that preceded the push."
    )

    def __call__(self):
        self._execute_or_complain(self.COMMAND)  # sets success
        if self.success:
            # Erase history so all steps before the push won't be rolled back
            self.log.debug("GitPush successful; erasing rollback history.")
            # self.releaser.rewindable.clear()
            self.releaser.rewindable[:] = []  # Python 2.6 has no clear()
            # self.releaser.non_rewindable.clear()
            self.releaser.non_rewindable[:] = []  # Python 2.6 has no clear()


class GitPushTags(CommandStep):
    """Pushes local tags to the remote repository. Can rollback()."""

    COMMAND = "git push --tags"
    ERROR_CODE = 56
    stop_on_failure = False

    def rollback(self):
        for tag in self.releaser.created_tags:
            self._execute_or_complain('git push --delete origin "v{0}"'.format(tag))



#!/usr/bin/env python3

"""Installer for the *releaser* utility."""

import os
from setuptools import setup, find_packages

# http://peak.telecommunity.com/DevCenter/setuptools#developer-s-guide


def content_of(*files):
    """Given 1 or more file paths, return their content in a single string."""
    here = os.path.abspath(os.path.dirname(__file__))
    content = []
    for f in files:
        with open(os.path.join(here, f), encoding="utf-8") as stream:
            content.append(stream.read())
    return "\n".join(content)


dependencies = [  # TODO UPGRADE THEM
    "bag > 0.3.7",
    "docutils < 0.13a",
    "grimace == 0.0.14",
    "path.py < 10.0a",  # TODO Switch to pathlib
    "requests",
    "wheel",
]

setup(
    name="releaser",
    description="Automates the process of releasing a new version of " "some software.",
    long_description=content_of("README.rst", "CHANGES.rst"),
    url="https://github.com/nandoflorestan/releaser",
    version="2.0.0",
    license="MIT",
    author="Nando Florestan",
    author_email="nandoflorestan@gmail.com",
    install_requires=dependencies,
    classifiers=[  # http://pypi.python.org/pypi?:action=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=[
        "python",
        "software",
        "release",
        "releaser",
        "automation",
        "git",
        "hg",
        "setup.py",
        "setuptools",
        "distribute",
    ],
    packages=find_packages(),
    include_package_data=True,
    # test_suite='releaser.tests',
    zip_safe=False,
)
